@@ -1,6 +1,7 @@
+/* SPDX-License-Identifier: BSD-2-Clause */
 /*
  * dhcpcd - DHCP client daemon
- * Copyright (c) 2006-2018 Roy Marples <roy@marples.name>
+ * Copyright (c) 2006-2021 Roy Marples <roy@marples.name>
  * All rights reserved
 
  * Redistribution and use in source and binary forms, with or without
@@ -30,6 +31,8 @@
 #include <sys/ioctl.h>
 #include <sys/socket.h>
 
+#include <fcntl.h> /* Needs to be here for old Linux */
+
 #include "config.h"
 
 #include <net/if.h>
@@ -58,10 +61,12 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <syslog.h>
 #include <unistd.h>
-#include <fcntl.h>
 
+#define ELOOP_QUEUE	ELOOP_IF
 #include "common.h"
+#include "eloop.h"
 #include "dev.h"
 #include "dhcp.h"
 #include "dhcp6.h"
@@ -71,6 +76,7 @@
 #include "ipv4ll.h"
 #include "ipv6nd.h"
 #include "logerr.h"
+#include "privsep.h"
 
 void
 if_free(struct interface *ifp)
@@ -78,12 +84,20 @@ if_free(struct interface *ifp)
 
 	if (ifp == NULL)
 		return;
+#ifdef IPV4LL
 	ipv4ll_free(ifp);
+#endif
+#ifdef INET
 	dhcp_free(ifp);
 	ipv4_free(ifp);
+#endif
+#ifdef DHCP6
 	dhcp6_free(ifp);
+#endif
+#ifdef INET6
 	ipv6nd_free(ifp);
 	ipv6_free(ifp);
+#endif
 	rt_freeif(ifp);
 	free_options(ifp->ctx, ifp->options);
 	free(ifp);
@@ -96,16 +110,20 @@ if_opensockets(struct dhcpcd_ctx *ctx)
 	if (if_opensockets_os(ctx) == -1)
 		return -1;
 
-	/* We use this socket for some operations without INET. */
-	ctx->pf_inet_fd = xsocket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
-	if (ctx->pf_inet_fd == -1)
-		return -1;
-
 #ifdef IFLR_ACTIVE
 	ctx->pf_link_fd = xsocket(PF_LINK, SOCK_DGRAM | SOCK_CLOEXEC, 0);
 	if (ctx->pf_link_fd == -1)
 		return -1;
+#ifdef HAVE_CAPSICUM
+	if (ps_rights_limit_ioctl(ctx->pf_link_fd) == -1)
+		return -1;
 #endif
+#endif
+
+	/* We use this socket for some operations without INET. */
+	ctx->pf_inet_fd = xsocket(PF_INET, SOCK_DGRAM | SOCK_CLOEXEC, 0);
+	if (ctx->pf_inet_fd == -1)
+		return -1;
 
 	return 0;
 }
@@ -116,7 +134,7 @@ if_closesockets(struct dhcpcd_ctx *ctx)
 
 	if (ctx->pf_inet_fd != -1)
 		close(ctx->pf_inet_fd);
-#ifdef IFLR_ACTIVE
+#ifdef PF_LINK
 	if (ctx->pf_link_fd != -1)
 		close(ctx->pf_link_fd);
 #endif
@@ -128,54 +146,101 @@ if_closesockets(struct dhcpcd_ctx *ctx)
 }
 
 int
-if_carrier(struct interface *ifp)
+if_ioctl(struct dhcpcd_ctx *ctx, ioctl_request_t req, void *data, size_t len)
 {
-	int r;
-	struct ifreq ifr;
-#ifdef SIOCGIFMEDIA
-	struct ifmediareq ifmr;
-#endif
 
-	memset(&ifr, 0, sizeof(ifr));
-	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
-	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFFLAGS, &ifr) == -1)
-		return LINK_UNKNOWN;
-	ifp->flags = (unsigned int)ifr.ifr_flags;
-
-#ifdef SIOCGIFMEDIA
-	memset(&ifmr, 0, sizeof(ifmr));
-	strlcpy(ifmr.ifm_name, ifp->name, sizeof(ifmr.ifm_name));
-	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFMEDIA, &ifmr) != -1 &&
-	    ifmr.ifm_status & IFM_AVALID)
-		r = (ifmr.ifm_status & IFM_ACTIVE) ? LINK_UP : LINK_DOWN;
-	else
-		r = ifr.ifr_flags & IFF_RUNNING ? LINK_UP : LINK_UNKNOWN;
-#else
-	r = ifr.ifr_flags & IFF_RUNNING ? LINK_UP : LINK_DOWN;
+#ifdef PRIVSEP
+	if (ctx->options & DHCPCD_PRIVSEP)
+		return (int)ps_root_ioctl(ctx, req, data, len);
 #endif
-	return r;
+	return ioctl(ctx->pf_inet_fd, req, data, len);
 }
 
 int
-if_setflag(struct interface *ifp, short flag)
+if_getflags(struct interface *ifp)
 {
-	struct ifreq ifr;
-	int r;
+	struct ifreq ifr = { .ifr_flags = 0 };
 
-	memset(&ifr, 0, sizeof(ifr));
 	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
-	r = -1;
-	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFFLAGS, &ifr) == 0) {
-		if (flag == 0 || (ifr.ifr_flags & flag) == flag)
-			r = 0;
-		else {
-			ifr.ifr_flags |= flag;
-			if (ioctl(ifp->ctx->pf_inet_fd, SIOCSIFFLAGS, &ifr) ==0)
-				r = 0;
-		}
-		ifp->flags = (unsigned int)ifr.ifr_flags;
+	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFFLAGS, &ifr) == -1)
+		return -1;
+	ifp->flags = (unsigned int)ifr.ifr_flags;
+	return 0;
+}
+
+int
+if_setflag(struct interface *ifp, short setflag, short unsetflag)
+{
+	struct ifreq ifr = { .ifr_flags = 0 };
+	short oflags;
+
+	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
+	if (ioctl(ifp->ctx->pf_inet_fd, SIOCGIFFLAGS, &ifr) == -1)
+		return -1;
+
+	oflags = ifr.ifr_flags;
+	ifr.ifr_flags |= setflag;
+	ifr.ifr_flags &= (short)~unsetflag;
+	if (ifr.ifr_flags != oflags &&
+	    if_ioctl(ifp->ctx, SIOCSIFFLAGS, &ifr, sizeof(ifr)) == -1)
+		return -1;
+
+	/*
+	 * Do NOT set ifp->flags here.
+	 * We need to listen for flag updates from the kernel as they
+	 * need to sync with carrier.
+	 */
+	return 0;
+}
+
+bool
+if_is_link_up(const struct interface *ifp)
+{
+
+	return ifp->flags & IFF_UP &&
+	    (ifp->carrier != LINK_DOWN ||
+	     (ifp->options != NULL && !(ifp->options->options & DHCPCD_LINK)));
+}
+
+int
+if_randomisemac(struct interface *ifp)
+{
+	uint32_t randnum;
+	size_t hwlen = ifp->hwlen, rlen = 0;
+	uint8_t buf[HWADDR_LEN], *bp = buf, *rp = (uint8_t *)&randnum;
+	char sbuf[HWADDR_LEN * 3];
+	int retval;
+
+	if (hwlen == 0) {
+		errno = ENOTSUP;
+		return -1;
 	}
-	return r;
+	if (hwlen > sizeof(buf)) {
+		errno = ENOBUFS;
+		return -1;
+	}
+
+	for (; hwlen != 0; hwlen--) {
+		if (rlen == 0) {
+			randnum = arc4random();
+			rp = (uint8_t *)&randnum;
+			rlen = sizeof(randnum);
+		}
+		*bp++ = *rp++;
+		rlen--;
+	}
+
+	/* Unicast address and locally administered. */
+	buf[0] &= 0xFC;
+	buf[0] |= 0x02;
+
+	logdebugx("%s: hardware address randomised to %s",
+	    ifp->name,
+	    hwaddr_ntoa(buf, ifp->hwlen, sbuf, sizeof(sbuf)));
+	retval = if_setmac(ifp, buf, ifp->hwlen);
+	if (retval == 0)
+		memcpy(ifp->hwaddr, buf, ifp->hwlen);
+	return retval;
 }
 
 static int
@@ -240,8 +305,15 @@ if_learnaddrs(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 			addrflags = if_addrflags(ifp, &addr->sin_addr,
 			    ifa->ifa_name);
 			if (addrflags == -1) {
-				if (errno != EEXIST && errno != EADDRNOTAVAIL)
-					logerr("%s: if_addrflags", __func__);
+				if (errno != EEXIST && errno != EADDRNOTAVAIL) {
+					char dbuf[INET_ADDRSTRLEN];
+					const char *dbp;
+
+					dbp = inet_ntop(AF_INET, &addr->sin_addr,
+					    dbuf, sizeof(dbuf));
+					logerr("%s: if_addrflags: %s%%%s",
+					    __func__, dbp, ifp->name);
+				}
 				continue;
 			}
 #endif
@@ -254,6 +326,7 @@ if_learnaddrs(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 		case AF_INET6:
 			sin6 = (void *)ifa->ifa_addr;
 			net6 = (void *)ifa->ifa_netmask;
+
 #ifdef __KAME__
 			if (IN6_IS_ADDR_LINKLOCAL(&sin6->sin6_addr))
 				/* Remove the scope from the address */
@@ -264,8 +337,15 @@ if_learnaddrs(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 			addrflags = if_addrflags6(ifp, &sin6->sin6_addr,
 			    ifa->ifa_name);
 			if (addrflags == -1) {
-				if (errno != EEXIST && errno != EADDRNOTAVAIL)
-					logerr("%s: if_addrflags6", __func__);
+				if (errno != EEXIST && errno != EADDRNOTAVAIL) {
+					char dbuf[INET6_ADDRSTRLEN];
+					const char *dbp;
+
+					dbp = inet_ntop(AF_INET6, &sin6->sin6_addr,
+					    dbuf, sizeof(dbuf));
+					logerr("%s: if_addrflags6: %s%%%s",
+					    __func__, dbp, ifp->name);
+				}
 				continue;
 			}
 #endif
@@ -277,7 +357,12 @@ if_learnaddrs(struct dhcpcd_ctx *ctx, struct if_head *ifs,
 		}
 	}
 
-	freeifaddrs(*ifaddrs);
+#ifdef PRIVSEP_GETIFADDRS
+	if (IN_PRIVSEP(ctx))
+		free(*ifaddrs);
+	else
+#endif
+		freeifaddrs(*ifaddrs);
 	*ifaddrs = NULL;
 }
 
@@ -314,6 +399,44 @@ if_valid_hwaddr(const uint8_t *hwaddr, size_t hwlen)
 	return false;
 }
 
+#if defined(AF_PACKET) && !defined(AF_LINK)
+static unsigned int
+if_check_arphrd(struct interface *ifp, unsigned int active, bool if_noconf)
+{
+
+	switch(ifp->hwtype) {
+	case ARPHRD_ETHER:	/* FALLTHROUGH */
+	case ARPHRD_IEEE1394:	/* FALLTHROUGH */
+	case ARPHRD_INFINIBAND:	/* FALLTHROUGH */
+	case ARPHRD_NONE:	/* FALLTHROUGH */
+		break;
+	case ARPHRD_LOOPBACK:
+	case ARPHRD_PPP:
+		if (if_noconf && active) {
+			logdebugx("%s: ignoring due to interface type and"
+			    " no config",
+			    ifp->name);
+			active = IF_INACTIVE;
+		}
+		break;
+	default:
+		if (active) {
+			int i;
+
+			if (if_noconf)
+				active = IF_INACTIVE;
+			i = active ? LOG_WARNING : LOG_DEBUG;
+			logmessage(i, "%s: unsupported"
+			    " interface type 0x%.2x",
+			    ifp->name, ifp->hwtype);
+		}
+		break;
+	}
+
+	return active;
+}
+#endif
+
 struct if_head *
 if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
     int argc, char * const *argv)
@@ -324,30 +447,38 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 	struct if_head *ifs;
 	struct interface *ifp;
 	struct if_spec spec;
+	bool if_noconf;
 #ifdef AF_LINK
 	const struct sockaddr_dl *sdl;
-#ifdef SIOCGIFPRIORITY
-	struct ifreq ifr;
-#endif
 #ifdef IFLR_ACTIVE
-	struct if_laddrreq iflr;
+	struct if_laddrreq iflr = { .flags = IFLR_PREFIX };
 #endif
-
-#ifdef IFLR_ACTIVE
-	memset(&iflr, 0, sizeof(iflr));
-#endif
-#elif AF_PACKET
+#elif defined(AF_PACKET)
 	const struct sockaddr_ll *sll;
 #endif
+#if defined(SIOCGIFPRIORITY)
+	struct ifreq ifr;
+#endif
 
 	if ((ifs = malloc(sizeof(*ifs))) == NULL) {
 		logerr(__func__);
 		return NULL;
 	}
 	TAILQ_INIT(ifs);
+
+#ifdef PRIVSEP_GETIFADDRS
+	if (ctx->options & DHCPCD_PRIVSEP) {
+		if (ps_root_getifaddrs(ctx, ifaddrs) == -1) {
+			logerr("ps_root_getifaddrs");
+			free(ifs);
+			return NULL;
+		}
+	} else
+#endif
 	if (getifaddrs(ifaddrs) == -1) {
-		logerr(__func__);
-		goto out;
+		logerr("getifaddrs");
+		free(ifs);
+		return NULL;
 	}
 
 	for (ifa = *ifaddrs; ifa; ifa = ifa->ifa_next) {
@@ -355,7 +486,7 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 #ifdef AF_LINK
 			if (ifa->ifa_addr->sa_family != AF_LINK)
 				continue;
-#elif AF_PACKET
+#elif defined(AF_PACKET)
 			if (ifa->ifa_addr->sa_family != AF_PACKET)
 				continue;
 #endif
@@ -406,23 +537,30 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 
 #ifdef PLUGIN_DEV
 		/* Ensure that the interface name has settled */
-		if (!dev_initialized(ctx, spec.devname))
-			continue;
-#endif
-
-		/* Don't allow loopback or pointopoint unless explicit */
-		if (ifa->ifa_flags & (IFF_LOOPBACK | IFF_POINTOPOINT)) {
-			if ((argc == 0 || argc == -1) &&
-			    ctx->ifac == 0 && !if_hasconf(ctx, spec.devname))
-				active = IF_INACTIVE;
-		}
-
-		if (if_vimaster(ctx, spec.devname) == 1) {
-			logfunc_t *logfunc = argc != 0 ? logerrx : logdebugx;
-			logfunc("%s: is a Virtual Interface Master, skipping",
+		if (!dev_initialised(ctx, spec.devname)) {
+			logdebugx("%s: waiting for interface to initialise",
 			    spec.devname);
 			continue;
 		}
+#endif
+
+		if (if_vimaster(ctx, spec.devname) == 1) {
+			int loglevel = argc != 0 ? LOG_ERR : LOG_DEBUG;
+			logmessage(loglevel,
+			    "%s: is a Virtual Interface Master, skipping",
+			    spec.devname);
+			continue;
+		}
+
+		if_noconf = ((argc == 0 || argc == -1) && ctx->ifac == 0 &&
+		    !if_hasconf(ctx, spec.devname));
+
+		/* Don't allow some reserved interface names unless explicit. */
+		if (if_noconf && if_ignore(ctx, spec.devname)) {
+			logdebugx("%s: ignoring due to interface type and"
+			    " no config", spec.devname);
+			active = IF_INACTIVE;
+		}
 
 		ifp = calloc(1, sizeof(*ifp));
 		if (ifp == NULL) {
@@ -458,26 +596,24 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 #ifdef IFT_BRIDGE
 			case IFT_BRIDGE: /* FALLTHROUGH */
 #endif
-#ifdef IFT_PPP
-			case IFT_PPP: /* FALLTHROUGH */
-#endif
 #ifdef IFT_PROPVIRTUAL
 			case IFT_PROPVIRTUAL: /* FALLTHROUGH */
 #endif
-#if defined(IFT_BRIDGE) || defined(IFT_PPP) || defined(IFT_PROPVIRTUAL)
+#ifdef IFT_TUNNEL
+			case IFT_TUNNEL: /* FALLTHROUGH */
+#endif
+			case IFT_LOOP: /* FALLTHROUGH */
+			case IFT_PPP:
 				/* Don't allow unless explicit */
-				if ((argc == 0 || argc == -1) &&
-				    ctx->ifac == 0 && active &&
-				    !if_hasconf(ctx, ifp->name))
-				{
+				if (if_noconf && active) {
 					logdebugx("%s: ignoring due to"
 					    " interface type and"
 					    " no config",
 					    ifp->name);
 					active = IF_INACTIVE;
 				}
+				__fallthrough; /* appease gcc */
 				/* FALLTHROUGH */
-#endif
 #ifdef IFT_L2VLAN
 			case IFT_L2VLAN: /* FALLTHROUGH */
 #endif
@@ -485,81 +621,62 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 			case IFT_L3IPVLAN: /* FALLTHROUGH */
 #endif
 			case IFT_ETHER:
-				ifp->family = ARPHRD_ETHER;
+				ifp->hwtype = ARPHRD_ETHER;
 				break;
 #ifdef IFT_IEEE1394
 			case IFT_IEEE1394:
-				ifp->family = ARPHRD_IEEE1394;
+				ifp->hwtype = ARPHRD_IEEE1394;
 				break;
 #endif
 #ifdef IFT_INFINIBAND
 			case IFT_INFINIBAND:
-				ifp->family = ARPHRD_INFINIBAND;
+				ifp->hwtype = ARPHRD_INFINIBAND;
 				break;
 #endif
 			default:
 				/* Don't allow unless explicit */
-				if ((argc == 0 || argc == -1) &&
-				    ctx->ifac == 0 &&
-				    !if_hasconf(ctx, ifp->name))
-					active = IF_INACTIVE;
-				if (active)
-					logwarnx("%s: unsupported"
-					    " interface type %.2x",
+				if (active) {
+					if (if_noconf)
+						active = IF_INACTIVE;
+					i = active ? LOG_WARNING : LOG_DEBUG;
+					logmessage(i, "%s: unsupported"
+					    " interface type 0x%.2x",
 					    ifp->name, sdl->sdl_type);
+				}
 				/* Pretend it's ethernet */
-				ifp->family = ARPHRD_ETHER;
+				ifp->hwtype = ARPHRD_ETHER;
 				break;
 			}
 			ifp->hwlen = sdl->sdl_alen;
 			memcpy(ifp->hwaddr, CLLADDR(sdl), ifp->hwlen);
-#elif AF_PACKET
+#elif defined(AF_PACKET)
 			sll = (const void *)ifa->ifa_addr;
 			ifp->index = (unsigned int)sll->sll_ifindex;
-			ifp->family = sll->sll_hatype;
+			ifp->hwtype = sll->sll_hatype;
 			ifp->hwlen = sll->sll_halen;
 			if (ifp->hwlen != 0)
 				memcpy(ifp->hwaddr, sll->sll_addr, ifp->hwlen);
+			active = if_check_arphrd(ifp, active, if_noconf);
 #endif
 		}
 #ifdef __linux__
-		/* PPP addresses on Linux don't have hardware addresses */
-		else
-			ifp->index = if_nametoindex(ifp->name);
-#endif
+		else {
+			struct ifreq ifr = { .ifr_flags = 0 };
 
-		/* Ensure hardware address is valid. */
-		if (!if_valid_hwaddr(ifp->hwaddr, ifp->hwlen))
-			ifp->hwlen = 0;
-
-		/* We only work on ethernet by default */
-		if (ifp->family != ARPHRD_ETHER) {
-			if ((argc == 0 || argc == -1) &&
-			    ctx->ifac == 0 && !if_hasconf(ctx, ifp->name))
-				active = IF_INACTIVE;
-			switch (ifp->family) {
-			case ARPHRD_IEEE1394:
-			case ARPHRD_INFINIBAND:
-#ifdef ARPHRD_LOOPBACK
-			case ARPHRD_LOOPBACK:
-#endif
-#ifdef ARPHRD_PPP
-			case ARPHRD_PPP:
-#endif
-				/* We don't warn for supported families */
-				break;
-
-/* IFT already checked */
-#ifndef AF_LINK
-			default:
-				if (active)
-					logwarnx("%s: unsupported"
-					    " interface family %.2x",
-					    ifp->name, ifp->family);
-				break;
-#endif
-			}
+			/* This is a huge bug in getifaddrs(3) as there
+			 * is no reason why this can't be returned in
+			 * ifa_addr. */
+			strlcpy(ifr.ifr_name, ifa->ifa_name,
+			    sizeof(ifr.ifr_name));
+			if (ioctl(ctx->pf_inet_fd, SIOCGIFHWADDR, &ifr) == -1)
+				logerr("%s: SIOCGIFHWADDR", ifa->ifa_name);
+			ifp->hwtype = ifr.ifr_hwaddr.sa_family;
+			if (ioctl(ctx->pf_inet_fd, SIOCGIFINDEX, &ifr) == -1)
+				logerr("%s: SIOCGIFINDEX", ifa->ifa_name);
+			ifp->index = (unsigned int)ifr.ifr_ifindex;
+			if_check_arphrd(ifp, active, if_noconf);
 		}
+#endif
 
 		if (!(ctx->options & (DHCPCD_DUMPLEASE | DHCPCD_TEST))) {
 			/* Handle any platform init for the interface */
@@ -576,36 +693,38 @@ if_discover(struct dhcpcd_ctx *ctx, struct ifaddrs **ifaddrs,
 		/* Respect the interface priority */
 		memset(&ifr, 0, sizeof(ifr));
 		strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
-		if (ioctl(ctx->pf_inet_fd, SIOCGIFPRIORITY, &ifr) == 0)
+		if (pioctl(ctx, SIOCGIFPRIORITY, &ifr, sizeof(ifr)) == 0)
 			ifp->metric = (unsigned int)ifr.ifr_metric;
 		if_getssid(ifp);
 #else
-		/* We reserve the 100 range for virtual interfaces, if and when
-		 * we can work them out. */
-		ifp->metric = 200 + ifp->index;
+		/* Leave a low portion for user config */
+		ifp->metric = RTMETRIC_BASE + ifp->index;
 		if (if_getssid(ifp) != -1) {
-			ifp->wireless = 1;
-			ifp->metric += 100;
+			ifp->wireless = true;
+			ifp->metric += RTMETRIC_WIRELESS;
 		}
 #endif
 
 		ifp->active = active;
-		if (ifp->active)
-			ifp->carrier = if_carrier(ifp);
-		else
-			ifp->carrier = LINK_UNKNOWN;
+		ifp->carrier = if_carrier(ifp, ifa->ifa_data);
 		TAILQ_INSERT_TAIL(ifs, ifp, next);
 	}
 
-out:
 	return ifs;
 }
 
-/* Decode bge0:1 as dev = bge, ppa = 0 and lun = 1 */
+/*
+ * eth0.100:2 OR eth0i100:2 (seems to be NetBSD xvif(4) only)
+ *
+ * drvname == eth
+ * devname == eth0.100 OR eth0i100
+ * ppa = 0
+ * lun = 2
+ */
 int
 if_nametospec(const char *ifname, struct if_spec *spec)
 {
-	char *ep;
+	char *ep, *pp;
 	int e;
 
 	if (ifname == NULL || *ifname == '\0' ||
@@ -617,6 +736,8 @@ if_nametospec(const char *ifname, struct if_spec *spec)
 		errno = EINVAL;
 		return -1;
 	}
+
+	/* :N is an alias */
 	ep = strchr(spec->drvname, ':');
 	if (ep) {
 		spec->lun = (int)strtoi(ep + 1, NULL, 10, 0, INT_MAX, &e);
@@ -624,23 +745,51 @@ if_nametospec(const char *ifname, struct if_spec *spec)
 			errno = e;
 			return -1;
 		}
-		*ep-- = '\0';
+		*ep = '\0';
+#ifdef __sun
+		ep--;
+#endif
 	} else {
 		spec->lun = -1;
+#ifdef __sun
 		ep = spec->drvname + strlen(spec->drvname) - 1;
+#endif
 	}
+
 	strlcpy(spec->devname, spec->drvname, sizeof(spec->devname));
+#ifdef __sun
+	/* Solaris has numbers in the driver name, such as e1000g */
 	while (ep > spec->drvname && isdigit((int)*ep))
 		ep--;
 	if (*ep++ == ':') {
 		errno = EINVAL;
 		return -1;
 	}
-	spec->ppa = (int)strtoi(ep, NULL, 10, 0, INT_MAX, &e);
-	if (e != 0)
-		spec->ppa = -1;
+#else
+	/* BSD and Linux no not have numbers in the driver name */
+	for (ep = spec->drvname; *ep != '\0' && !isdigit((int)*ep); ep++) {
+		if (*ep == ':') {
+			errno = EINVAL;
+			return -1;
+		}
+	}
+#endif
+	spec->ppa = (int)strtoi(ep, &pp, 10, 0, INT_MAX, &e);
 	*ep = '\0';
 
+#ifndef __sun
+	/*
+	 * . is used for VLAN style names
+	 * i is used on NetBSD for xvif interfaces
+	 */
+	if (pp != NULL && (*pp == '.' || *pp == 'i')) {
+		spec->vlid = (int)strtoi(pp + 1, NULL, 10, 0, INT_MAX, &e);
+		if (e)
+			spec->vlid = -1;
+	} else
+#endif
+		spec->vlid = -1;
+
 	return 0;
 }
 
@@ -698,89 +847,111 @@ if_domtu(const struct interface *ifp, short int mtu)
 	int r;
 	struct ifreq ifr;
 
+#ifdef __sun
+	if (mtu == 0)
+		return if_mtu_os(ifp);
+#endif
+
 	memset(&ifr, 0, sizeof(ifr));
 	strlcpy(ifr.ifr_name, ifp->name, sizeof(ifr.ifr_name));
 	ifr.ifr_mtu = mtu;
-	r = ioctl(ifp->ctx->pf_inet_fd, mtu ? SIOCSIFMTU : SIOCGIFMTU, &ifr);
+	if (mtu != 0)
+		r = if_ioctl(ifp->ctx, SIOCSIFMTU, &ifr, sizeof(ifr));
+	else
+		r = pioctl(ifp->ctx, SIOCGIFMTU, &ifr, sizeof(ifr));
+
 	if (r == -1)
 		return -1;
 	return ifr.ifr_mtu;
 }
 
-/* Interface comparer for working out ordering. */
-static int
-if_cmp(const struct interface *si, const struct interface *ti)
+#ifdef ALIAS_ADDR
+int
+if_makealias(char *alias, size_t alias_len, const char *ifname, int lun)
 {
-#ifdef INET
-	int r;
-#endif
 
-	/* Check active first */
-	if (si->active > ti->active)
-		return -1;
-	if (si->active < ti->active)
-		return 1;
-
-	/* Check carrier status next */
-	if (si->carrier > ti->carrier)
-		return -1;
-	if (si->carrier < ti->carrier)
-		return 1;
-
-	if (D_STATE_RUNNING(si) && !D_STATE_RUNNING(ti))
-		return -1;
-	if (!D_STATE_RUNNING(si) && D_STATE_RUNNING(ti))
-		return 1;
-	if (RS_STATE_RUNNING(si) && !RS_STATE_RUNNING(ti))
-		return -1;
-	if (!RS_STATE_RUNNING(si) && RS_STATE_RUNNING(ti))
-		return 1;
-	if (D6_STATE_RUNNING(si) && !D6_STATE_RUNNING(ti))
-		return -1;
-	if (!D6_STATE_RUNNING(si) && D6_STATE_RUNNING(ti))
-		return 1;
-
-#ifdef INET
-	/* Special attention needed here due to states and IPv4LL. */
-	if ((r = ipv4_ifcmp(si, ti)) != 0)
-		return r;
-#endif
-
-	/* Finally, metric */
-	if (si->metric < ti->metric)
-		return -1;
-	if (si->metric > ti->metric)
-		return 1;
-	return 0;
+	if (lun == 0)
+		return strlcpy(alias, ifname, alias_len);
+	return snprintf(alias, alias_len, "%s:%u", ifname, lun);
 }
+#endif
 
-/* Sort the interfaces into a preferred order - best first, worst last. */
-void
-if_sortinterfaces(struct dhcpcd_ctx *ctx)
+struct interface *
+if_findifpfromcmsg(struct dhcpcd_ctx *ctx, struct msghdr *msg, int *hoplimit)
 {
-	struct if_head sorted;
-	struct interface *ifp, *ift;
+	struct cmsghdr *cm;
+	unsigned int ifindex = 0;
+	struct interface *ifp;
+#ifdef INET
+#ifdef IP_RECVIF
+	struct sockaddr_dl sdl;
+#else
+	struct in_pktinfo ipi;
+#endif
+#endif
+#ifdef INET6
+	struct in6_pktinfo ipi6;
+#else
+	UNUSED(hoplimit);
+#endif
 
-	if (ctx->ifaces == NULL ||
-	    (ifp = TAILQ_FIRST(ctx->ifaces)) == NULL ||
-	    TAILQ_NEXT(ifp, next) == NULL)
-		return;
-
-	TAILQ_INIT(&sorted);
-	TAILQ_REMOVE(ctx->ifaces, ifp, next);
-	TAILQ_INSERT_HEAD(&sorted, ifp, next);
-	while ((ifp = TAILQ_FIRST(ctx->ifaces))) {
-		TAILQ_REMOVE(ctx->ifaces, ifp, next);
-		TAILQ_FOREACH(ift, &sorted, next) {
-			if (if_cmp(ifp, ift) == -1) {
-				TAILQ_INSERT_BEFORE(ift, ifp, next);
+	for (cm = (struct cmsghdr *)CMSG_FIRSTHDR(msg);
+	     cm;
+	     cm = (struct cmsghdr *)CMSG_NXTHDR(msg, cm))
+	{
+#ifdef INET
+		if (cm->cmsg_level == IPPROTO_IP) {
+			switch(cm->cmsg_type) {
+#ifdef IP_RECVIF
+			case IP_RECVIF:
+				if (cm->cmsg_len <
+				    offsetof(struct sockaddr_dl, sdl_index) +
+				    sizeof(sdl.sdl_index))
+					continue;
+				memcpy(&sdl, CMSG_DATA(cm),
+				    MIN(sizeof(sdl), cm->cmsg_len));
+				ifindex = sdl.sdl_index;
+				break;
+#else
+			case IP_PKTINFO:
+				if (cm->cmsg_len != CMSG_LEN(sizeof(ipi)))
+					continue;
+				memcpy(&ipi, CMSG_DATA(cm), sizeof(ipi));
+				ifindex = (unsigned int)ipi.ipi_ifindex;
+				break;
+#endif
+			}
+		}
+#endif
+#ifdef INET6
+		if (cm->cmsg_level == IPPROTO_IPV6) {
+			switch(cm->cmsg_type) {
+			case IPV6_PKTINFO:
+				if (cm->cmsg_len != CMSG_LEN(sizeof(ipi6)))
+					continue;
+				memcpy(&ipi6, CMSG_DATA(cm), sizeof(ipi6));
+				ifindex = (unsigned int)ipi6.ipi6_ifindex;
+				break;
+			case IPV6_HOPLIMIT:
+				if (cm->cmsg_len != CMSG_LEN(sizeof(int)))
+					continue;
+				if (hoplimit == NULL)
+					break;
+				memcpy(hoplimit, CMSG_DATA(cm), sizeof(int));
 				break;
 			}
 		}
-		if (ift == NULL)
-			TAILQ_INSERT_TAIL(&sorted, ifp, next);
+#endif
 	}
-	TAILQ_CONCAT(ctx->ifaces, &sorted, next);
+
+	/* Find the receiving interface */
+	TAILQ_FOREACH(ifp, ctx->ifaces, next) {
+		if (ifp->index == ifindex)
+			break;
+	}
+	if (ifp == NULL)
+		errno = ESRCH;
+	return ifp;
 }
 
 int
@@ -822,3 +993,50 @@ out:
 	return -1;
 #endif
 }
+
+int
+xsocketpair(int domain, int type, int protocol, int fd[2])
+{
+	int s;
+#if !defined(HAVE_SOCK_CLOEXEC) || !defined(HAVE_SOCK_NONBLOCK)
+	int xflags, xtype = type;
+#endif
+
+#ifndef HAVE_SOCK_CLOEXEC
+	if (xtype & SOCK_CLOEXEC)
+		type &= ~SOCK_CLOEXEC;
+#endif
+#ifndef HAVE_SOCK_NONBLOCK
+	if (xtype & SOCK_NONBLOCK)
+		type &= ~SOCK_NONBLOCK;
+#endif
+
+	if ((s = socketpair(domain, type, protocol, fd)) == -1)
+		return -1;
+
+#ifndef HAVE_SOCK_CLOEXEC
+	if ((xtype & SOCK_CLOEXEC) && ((xflags = fcntl(fd[0], F_GETFD)) == -1 ||
+	    fcntl(fd[0], F_SETFD, xflags | FD_CLOEXEC) == -1))
+		goto out;
+	if ((xtype & SOCK_CLOEXEC) && ((xflags = fcntl(fd[1], F_GETFD)) == -1 ||
+	    fcntl(fd[1], F_SETFD, xflags | FD_CLOEXEC) == -1))
+		goto out;
+#endif
+#ifndef HAVE_SOCK_NONBLOCK
+	if ((xtype & SOCK_NONBLOCK) && ((xflags = fcntl(fd[0], F_GETFL)) == -1 ||
+	    fcntl(fd[0], F_SETFL, xflags | O_NONBLOCK) == -1))
+		goto out;
+	if ((xtype & SOCK_NONBLOCK) && ((xflags = fcntl(fd[1], F_GETFL)) == -1 ||
+	    fcntl(fd[1], F_SETFL, xflags | O_NONBLOCK) == -1))
+		goto out;
+#endif
+
+	return s;
+
+#if !defined(HAVE_SOCK_CLOEXEC) || !defined(HAVE_SOCK_NONBLOCK)
+out:
+	close(fd[0]);
+	close(fd[1]);
+	return -1;
+#endif
+}
